/********************************************************************************
** Form generated from reading UI file 'mainwindow.ui'
**
** Created by: Qt User Interface Compiler version 5.15.19
**
** WARNING! All changes made in this file will be lost when recompiling UI file!
********************************************************************************/

#ifndef UI_MAINWINDOW_H
#define UI_MAINWINDOW_H

#include <QtCore/QVariant>
#include <QtWidgets/QApplication>
#include <QtWidgets/QCheckBox>
#include <QtWidgets/QComboBox>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QLabel>
#include <QtWidgets/QMainWindow>
#include <QtWidgets/QSpacerItem>
#include <QtWidgets/QVBoxLayout>
#include <QtWidgets/QWidget>
#include "svgview.h"

QT_BEGIN_NAMESPACE

class Ui_MainWindow
{
public:
    QWidget *centralWidget;
    QVBoxLayout *verticalLayout_2;
    QHBoxLayout *horizontalLayout;
    QLabel *label_3;
    QComboBox *cmbBoxSize;
    QSpacerItem *horizontalSpacer;
    QLabel *label_5;
    QComboBox *cmbBoxBackground;
    QSpacerItem *horizontalSpacer_3;
    QCheckBox *chBoxDrawBorder;
    QSpacerItem *horizontalSpacer_2;
    SvgView *svgView;

    void setupUi(QMainWindow *MainWindow)
    {
        if (MainWindow->objectName().isEmpty())
            MainWindow->setObjectName(QString::fromUtf8("MainWindow"));
        MainWindow->resize(787, 424);
        centralWidget = new QWidget(MainWindow);
        centralWidget->setObjectName(QString::fromUtf8("centralWidget"));
        verticalLayout_2 = new QVBoxLayout(centralWidget);
        verticalLayout_2->setSpacing(6);
        verticalLayout_2->setContentsMargins(11, 11, 11, 11);
        verticalLayout_2->setObjectName(QString::fromUtf8("verticalLayout_2"));
        verticalLayout_2->setContentsMargins(0, 6, 0, 0);
        horizontalLayout = new QHBoxLayout();
        horizontalLayout->setSpacing(6);
        horizontalLayout->setObjectName(QString::fromUtf8("horizontalLayout"));
        horizontalLayout->setContentsMargins(2, -1, -1, -1);
        label_3 = new QLabel(centralWidget);
        label_3->setObjectName(QString::fromUtf8("label_3"));
        QSizePolicy sizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred);
        sizePolicy.setHorizontalStretch(0);
        sizePolicy.setVerticalStretch(0);
        sizePolicy.setHeightForWidth(label_3->sizePolicy().hasHeightForWidth());
        label_3->setSizePolicy(sizePolicy);

        horizontalLayout->addWidget(label_3);

        cmbBoxSize = new QComboBox(centralWidget);
        cmbBoxSize->addItem(QString());
        cmbBoxSize->addItem(QString());
        cmbBoxSize->setObjectName(QString::fromUtf8("cmbBoxSize"));

        horizontalLayout->addWidget(cmbBoxSize);

        horizontalSpacer = new QSpacerItem(20, 1, QSizePolicy::Fixed, QSizePolicy::Minimum);

        horizontalLayout->addItem(horizontalSpacer);

        label_5 = new QLabel(centralWidget);
        label_5->setObjectName(QString::fromUtf8("label_5"));

        horizontalLayout->addWidget(label_5);

        cmbBoxBackground = new QComboBox(centralWidget);
        cmbBoxBackground->addItem(QString());
        cmbBoxBackground->addItem(QString());
        cmbBoxBackground->addItem(QString());
        cmbBoxBackground->setObjectName(QString::fromUtf8("cmbBoxBackground"));

        horizontalLayout->addWidget(cmbBoxBackground);

        horizontalSpacer_3 = new QSpacerItem(20, 1, QSizePolicy::Fixed, QSizePolicy::Minimum);

        horizontalLayout->addItem(horizontalSpacer_3);

        chBoxDrawBorder = new QCheckBox(centralWidget);
        chBoxDrawBorder->setObjectName(QString::fromUtf8("chBoxDrawBorder"));

        horizontalLayout->addWidget(chBoxDrawBorder);

        horizontalSpacer_2 = new QSpacerItem(40, 20, QSizePolicy::Expanding, QSizePolicy::Minimum);

        horizontalLayout->addItem(horizontalSpacer_2);


        verticalLayout_2->addLayout(horizontalLayout);

        svgView = new SvgView(centralWidget);
        svgView->setObjectName(QString::fromUtf8("svgView"));
        QSizePolicy sizePolicy1(QSizePolicy::Preferred, QSizePolicy::Expanding);
        sizePolicy1.setHorizontalStretch(0);
        sizePolicy1.setVerticalStretch(0);
        sizePolicy1.setHeightForWidth(svgView->sizePolicy().hasHeightForWidth());
        svgView->setSizePolicy(sizePolicy1);

        verticalLayout_2->addWidget(svgView);

        MainWindow->setCentralWidget(centralWidget);

        retranslateUi(MainWindow);

        QMetaObject::connectSlotsByName(MainWindow);
    } // setupUi

    void retranslateUi(QMainWindow *MainWindow)
    {
        MainWindow->setWindowTitle(QCoreApplication::translate("MainWindow", "viewsvg", nullptr));
        label_3->setText(QCoreApplication::translate("MainWindow", "Size:", nullptr));
        cmbBoxSize->setItemText(0, QCoreApplication::translate("MainWindow", "Original", nullptr));
        cmbBoxSize->setItemText(1, QCoreApplication::translate("MainWindow", "Fit to View", nullptr));

        label_5->setText(QCoreApplication::translate("MainWindow", "Background:", nullptr));
        cmbBoxBackground->setItemText(0, QCoreApplication::translate("MainWindow", "None", nullptr));
        cmbBoxBackground->setItemText(1, QCoreApplication::translate("MainWindow", "White", nullptr));
        cmbBoxBackground->setItemText(2, QCoreApplication::translate("MainWindow", "Check board", nullptr));

        chBoxDrawBorder->setText(QCoreApplication::translate("MainWindow", "Draw image border", nullptr));
    } // retranslateUi

};

namespace Ui {
    class MainWindow: public Ui_MainWindow {};
} // namespace Ui

QT_END_NAMESPACE

#endif // UI_MAINWINDOW_H
